import os
import pygit2
import subprocess
import urllib.parse

from modules import packages, patch, utils


class GitError(Exception):
    pass

class GitFatalError(GitError):
    pass

class GitConfigError(GitError):
    def __init__(self, msg, suggested_command=None, git_command=None):
        super().__init__(msg)
        self.suggested_command = suggested_command
        self.git_command = git_command

class GitNoRemote(GitError):
    pass

class GitBadProject(GitError):
    pass


class Repo:
    def __init__(self, path):
        self.path = os.path.expanduser(path)
        try:
            self.repo = pygit2.Repository(self.path)
        except pygit2.GitError:
            raise GitFatalError('\'{}\' is not a git repository.'.format(path))
        self._config_cache = None

    @classmethod
    def clone(cls, remote, path, output_callback=None, check=True):
        cls._run_git_stream('clone', ['--progress', remote, path], None, output_callback,
                            check=check)
        return cls(path)

    @classmethod
    def _run_git_stream(cls, command, args, path, output_callback, check):
        """Runs an external git command `command` on the repo in `path`,
        passing `args` to it. `output_callback` receives the output in
        repeated calls until git finishes. Returns True on success. If
        `check` is True, raises a GitFatalError exception on failure."""
        if path:
            full_args = ['git', '-C', path, command]
        else:
            full_args = ['git', command]
        full_args.extend(args)
        result = utils.run_cb(full_args, output_callback, git=command)
        if check and result != 0:
            raise GitFatalError('Failed to run git. The command was: {}'
                                .format(' '.join(full_args)))
        return result == 0

    def _run_git(self, command, *args, check=False):
        """Runs an external git command `command`, passing `args` to it. If
        `check` is True and the command fails, a GitFatalError exception is
        raised. Returns stdout of the git command (as a string) or None if
        the command failed and `check` was False. Note the difference
        between returned "" and None."""
        full_args = ['git', '-C', self.path, command]
        full_args.extend(args)
        result = utils.run(full_args, git=command, check=False, universal_newlines=True,
                           stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                           errors='replace')
        if result.returncode == 0:
            return result.stdout
        if check:
            raise GitFatalError('Failed to run git. The command was: {}'
                                .format(' '.join(full_args)))
        return None

    def _config(self):
        # Unfortunatelly, we cannot parse the config here. The output from
        # git config list is ambiguous. First, '=' is a permitted part of
        # a subsection name, as well as of the value; things like
        # "remote.weird=.url=https://server/=x" are possible. Second, '.' is
        # also a permitted part of a subsection name; meaning we cannot
        # split the name to individual parts.
        # Ideally, we would store the result in a trie for fast lookups. But
        # configs tend to be short and it doesn't justify bringing in
        # another 3rd party dependency.
        if self._config_cache is None:
            self._config_cache = utils.splitlines(self._run_git('config', 'list', check=True))
        return self._config_cache

    def config_value(self, *args, all=False):
        """Lookup the given value in the git config. Each argument is a
        section, a subsection or a key, forming a path to look up. If `all`
        is False, returns the last value in the config or None if it's not
        present. If `all` is True, returns a list of all values."""
        result = []
        lookup = '.'.join(args) + '='
        for entry in self._config():
            if entry.startswith(lookup):
                result.append(entry[len(lookup):])
        if all:
            return result
        if not result:
            return None
        return result[-1]

    def config_values(self, *args):
        return self.config_value(*args, all=True)

    def fetch(self, remote_name, output_callback=None, commit=None, check=True):
        args = ['--progress', remote_name]
        if commit is not None:
            args.append(commit)
        return self._run_git_stream('fetch', args, self.path, output_callback, check=check)

    def full_sha(self, revision):
        return str(self.repo.revparse_single(revision).id)

    def get_diff(self, commit, silent=False):
        if not isinstance(commit, str):
            commit = str(commit.id)
        result = self._run_git('show', '-p', commit)
        if result is None and not silent:
            raise KeyError('commit {} not present in {}'.format(commit, self.path))
        return result or ''

    def remotes(self):
        result = []
        for name in utils.splitlines(self._run_git('remote', check=True)):
            url = self.config_value('remote', name, 'url')
            refspecs = self.config_values('remote', name, 'fetch')
            result.append(Remote(name, url, refspecs))
        return result


class Remote:
    def __init__(self, name, url, refspecs):
        self.name = name
        self.url = url
        self.refspecs = refspecs


class ProjectRepo(Repo):
    def get_all_projects(self, labname):
        result = []
        for remote in self.remotes():
            if remote.url.startswith('http'):
                parsed = urllib.parse.urlsplit(remote.url)
                server = parsed.netloc
                project = parsed.path
                if project.startswith('/'):
                    project = project[1:]
            else:
                try:
                    server, project = remote.url.split(':')
                    login, server = server.split('@')
                except ValueError:
                    continue
            if project.endswith('/'):
                project = project[:-1]
            if project.endswith('.git'):
                project = project[:-4]
            if server != labname:
                continue
            result.append((project, remote))
        return result

    def _get_mr_refs(self, labname, project_filter):
        """Returns two lists: a list of dictionaries with merge request
        refspecs, and a list of remotes (as a tuple (project, remote)) that
        do not have a merge request refspec set up. The project_filter is
        a function that gets the project name and should return True/False
        to consider/skip this remote."""
        mr_refs = []
        no_mr_ref = []
        for project, remote in self.get_all_projects(labname):
            if not project_filter(project):
                continue
            found = False
            for ref in remote.refspecs:
                if ref.startswith('^'):
                    # Ignore negative refspecs. This is not entirely
                    # correct, as we should check whether the refspec we
                    # care about is not negated. Let's ignore this until
                    # someone complains.
                    continue
                if ref.startswith('+'):
                    ref = ref[1:]
                src, dst = ref.split(':')
                if src.startswith('refs/merge-requests/') and dst.endswith('/*'):
                    mr_refs.append({ 'remote': remote,
                                     'project': project,
                                     'short_project': project.split('/')[-1],
                                     'pkg': packages.find(project),
                                     'ref': dst[:-2] })
                    found = True
            if not found:
                no_mr_ref.append((project, remote))
        return mr_refs, no_mr_ref

    def _store_mr_ref(self, mr_ref):
        for k, v in mr_ref.items():
            setattr(self, k, v)

    def find_remote(self, labname, projectname):
        pkg = packages.find(projectname)
        if not pkg:
            raise GitBadProject('Unsupported project: {}'.format(projectname))

        mr_refs, no_mr_ref = self._get_mr_refs(labname,
                                               lambda project: project == projectname)
        if len(mr_refs) == 1:
            self._store_mr_ref(mr_refs[0])
            return self.pkg
        if len(mr_refs) > 1 or len(no_mr_ref) > 1:
            raise GitNoRemote('Multiple remotes matching {} found in repository \'{}\'.'
                              .format(projectname, self.path))
        if len(no_mr_ref) == 0:
            raise GitNoRemote('No remote matching {} found in repository \'{}\'.'
                              .format(projectname, self.path))

        # There's no merge requests refspec but we've got the right project
        # and remote. Store them.
        self.project = projectname
        self.remote = no_mr_ref[0][1]

        cmd = ['git']
        if self.path != '.':
            cmd.extend(('-C', self.path))
        cmd.extend(('config', '--add', 'remote.{}.fetch'.format(self.remote.name),
                    '+refs/merge-requests/*/head:refs/remotes/{}/merge-requests/*'.format(self.remote.name)))
        raise GitConfigError('No refspec with merge requests found. Please run:\n' + ' '.join(cmd),
                             cmd, 'config')

    def find_projects(self, labname):
        mr_refs, no_mr_ref = self._get_mr_refs(labname,
                                               lambda project: packages.find(project))
        result = [m['project'] for m in mr_refs]
        result.extend(m[0] for m in no_mr_ref)
        if not result:
            raise GitNoRemote('No remote with a downstream project found in repository \'{}\'.'
                              .format(self.path))
        result.sort()
        return result

    def fetch(self, output_callback=None, commit=None, check=True):
        return super().fetch(self.remote.name, output_callback, commit, check)

    def get_mr_head(self, mr_id):
        return self.full_sha('{}/{}'.format(self.ref, mr_id))

    def get_commits(self, output_callback, base, head):
        try:
            head = self.repo.revparse_single(head).id
        except KeyError:
            # try to fetch the head in case it is a historic commit that was
            # not fetched by the standard 'git fetch'
            self.fetch(output_callback, commit=head, check=False)
            head = self.repo.revparse_single(head).id
        base = self.repo.revparse_single(base).id
        walker = self.repo.walk(head)
        walker.sort(pygit2.GIT_SORT_TOPOLOGICAL | pygit2.GIT_SORT_REVERSE)
        walker.hide(base)
        return walker

    def get_diff_range(self, base, head):
        args = ('git', '-C', self.path, 'diff', '{}..{}'.format(base, head))
        for i in range(2):
            result = utils.run(args, git='diff', check=False, universal_newlines=True,
                               stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                               errors='replace')
            if result.returncode == 0:
                return result.stdout
            # try to fetch the head in case it is a historic commit that was
            # not fetched by the standard 'git fetch'
            if i > 0 or not self.fetch(commit=head, check=False):
                result.check_returncode()
                assert False

    def get_parsed_diff(self, sha):
        return patch.PatchParser(sha[:12], self.get_diff(sha)).parse()

    def get_parsed_diff_range(self, base, head):
        return patch.PatchParser('{:.12}..{:.12}'.format(base, head),
                                 self.get_diff_range(base, head)).parse()


class RepoCollection:
    def __init__(self, paths, labname):
        self.repos = [self._new_repo(path) for path in paths]
        self.labname = labname

    def _new_repo(self, path):
        return ProjectRepo(path)

    def add_cur_dir(self):
        """Adds the current directory (.) as the first repo in the
        collection. Returns the available project names or raises a GitError
        exception."""
        repo = self._new_repo('.')
        results = repo.find_projects(self.labname)
        self.repos.insert(0, repo)
        return results

    def add_repo(self, path, check_projectname):
        repo = self._new_repo(path)
        repo.find_remote(self.labname, check_projectname)
        self.repos.append(repo)

    def find_repo(self, projectname):
        for repo in self.repos:
            try:
                pkg = repo.find_remote(self.labname, projectname)
                return repo, pkg
            except GitNoRemote:
                pass
        return None, None

    def get_commit(self, commit_id):
        for repo in self.repos:
            try:
                return repo.repo[commit_id]
            except KeyError:
                continue
        return None

    def get_diff(self, commit):
        if not isinstance(commit, str):
            commit = str(commit.id)
        for repo in self.repos:
            if not repo.repo.get(commit):
                continue
            diff = repo.get_diff(commit, silent=True)
            if diff:
                return diff
        return ''
